package be.iminds.ilabt.jfed.experimenter_gui.editor.ontology;

import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.controller.NodePropertyOverview;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.model.query.SparqlService;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.StageUtils;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import com.lynden.gmapsfx.ClusteredGoogleMapView;
import com.lynden.gmapsfx.MapComponentInitializedListener;
import com.lynden.gmapsfx.MapReadyListener;
import com.lynden.gmapsfx.javascript.event.MapStateEventType;
import com.lynden.gmapsfx.javascript.event.UIEventType;
import com.lynden.gmapsfx.javascript.object.*;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.collections.transformation.FilteredList;
import javafx.fxml.FXML;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.ComboBox;
import javafx.scene.layout.BorderPane;
import javafx.stage.Stage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * User: twalcari
 * Date: 1/8/2015
 * Time: 14:34
 */
public class ResultMapDialog extends BorderPane implements MapReadyListener, MapComponentInitializedListener {
    private static final Logger LOG = LoggerFactory.getLogger(ResultMapDialog.class);
    @FXML
    private ClusteredGoogleMapView mapView;

    @FXML
    private ComboBox<SparqlService.NodeInfo> nodesComboBox;

    @FXML
    private Button selectNodeButton;

    private boolean mapViewIsInitialized = false;
    private ClusteredGoogleMap map;

    private final ObservableList<SparqlService.NodeInfo> nodes = FXCollections.observableArrayList();
    private final FilteredList<SparqlService.NodeInfo> nodesWithinBounds = new FilteredList<SparqlService.NodeInfo>(nodes);

    private Map<SparqlService.NodeInfo, LatLong> nodeLocations = new HashMap<>();


    private ObjectProperty<SparqlService.NodeInfo> result = new SimpleObjectProperty<>(null);
//    private Marker selectedLocationMarker;

    //dialog variables
    private final Stage dialogStage;

    public ResultMapDialog(){
        FXMLUtil.injectFXML(this);

        mapView.addMapInializedListener(this);
        mapView.addMapReadyListener(this);

        nodesComboBox.setConverter(NodePropertyOverview.NODE_INFO_STRING_CONVERTER);
        selectNodeButton.disableProperty().bind(nodesComboBox.valueProperty().isNull());

        nodesComboBox.setOnAction(event -> {
            SparqlService.NodeInfo node = nodesComboBox.getValue();
            if (node != null) {
                assert (node.getLongitude() != null && node.getLongitude() != null);
                map.panTo(nodeLocations.get(node));
                map.setZoom(18);
            }
        });



        Scene scene = new Scene(this);

        dialogStage = new Stage();
        StageUtils.setJFedStageIcons(dialogStage);
        dialogStage.setScene(scene);

        dialogStage.setTitle("Map of query results");
        dialogStage.setResizable(true);
    }

    @Override
    public void mapReady() {
        addNodesToMap();
    }

    @Override
    public void mapInitialized() {
        mapViewIsInitialized = true;

        tryInitializeMap();
    }

    private static final class ComparableLatLong {
        private final double lat, lon;

        public ComparableLatLong(double lat, double lon) {
            this.lon = lon;
            this.lat = lat;
        }

        public double getLat() {
            return lat;
        }

        public double getLon() {
            return lon;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ComparableLatLong that = (ComparableLatLong) o;

            if (Double.compare(that.lat, lat) != 0) return false;
            if (Double.compare(that.lon, lon) != 0) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result;
            long temp;
            temp = Double.doubleToLongBits(lat);
            result = (int) (temp ^ (temp >>> 32));
            temp = Double.doubleToLongBits(lon);
            result = 31 * result + (int) (temp ^ (temp >>> 32));
            return result;
        }

        public LatLong toLatLong() {
            return new LatLong(lat, lon);
        }
    }


    private void tryInitializeMap() {
        if (!mapViewIsInitialized) return;

        MapOptions mapOptions = new MapOptions();

        mapOptions.mapType(MapTypeIdEnum.ROADMAP)
                .center(new LatLong(41.711, 26.828))
                .overviewMapControl(true)
                .panControl(true)
                .rotateControl(true)
                .scaleControl(true)
                .streetViewControl(true)
                .zoomControl(true)
                .zoom(3)
                .maxZoom(18);

        map = mapView.createMap(mapOptions);
        map.getMarkerClusterer().setMaxZoom(17);
        map.getMarkerClusterer().setGridSize(20);

        map.addStateEventHandler(MapStateEventType.bounds_changed, this::onMapBoundsChanged);
        map.addUIEventHandler(UIEventType.click, obj -> {
            clearNodeSelection();
        });
    }
    private void addNodesToMap(){
        Set<ComparableLatLong> uniqueLocations = new HashSet<>();

        nodeLocations.clear();
        map.getMarkerClusterer().clearMarkers();

        //first setup locations, but try to prevent duplicate locations by shifting them a bit if necessary
        nodes.stream().filter(node -> node.getLatitude() != null && node.getLongitude() != null)
                .forEach(node -> {

                            ComparableLatLong ll = new ComparableLatLong(node.getLatitude(), node.getLongitude());

                            boolean changeLat = true;
                            while (uniqueLocations.contains(ll)) {
                                if (changeLat)
                                    ll = new ComparableLatLong(ll.getLat() + 0.0001, ll.getLon());
                                else
                                    ll = new ComparableLatLong(ll.getLat(), ll.getLon() + 0.0001);
                                changeLat = !changeLat;
                            }
                            uniqueLocations.add(ll);

                            nodeLocations.put(node, ll.toLatLong());
                        }


                );

        LatLongBounds bounds = new LatLongBounds();

        for (Map.Entry<SparqlService.NodeInfo, LatLong> entry : nodeLocations.entrySet()) {
            MarkerOptions nodeMarkerOptions = new MarkerOptions();
            nodeMarkerOptions.position(entry.getValue());
            nodeMarkerOptions.title(entry.getKey().getComponentUrn().getEncodedResourceName());

            Marker nodeMarker = new Marker(nodeMarkerOptions);

            LOG.trace("Adding marker {} at location {}", entry.getKey().getComponentManagerUrn().getEncodedResourceName(), entry.getValue());
            map.addClusterableMarker(nodeMarker);

            map.addUIEventHandler(nodeMarker, UIEventType.click, obj -> selectNode(entry.getKey()));


            bounds.extend(entry.getValue());
        }

    }

    private void onMapBoundsChanged() {
        final LatLongBounds newBounds = map.getBounds();

        nodesWithinBounds.setPredicate(node -> newBounds.contains(nodeLocations.get(node)));
    }


    private void clearNodeSelection() {
        nodesComboBox.setValue(null);
    }

    private void selectNode(SparqlService.NodeInfo node) {
        assert (node.getLatitude() != null && node.getLongitude() != null);
        map.panTo(nodeLocations.get(node));

        nodesComboBox.setValue(node);
    }

    public void setNodes(List<SparqlService.NodeInfo> nodes) {
        this.nodes.setAll(FXCollections.observableList(nodes));
        nodesComboBox.setItems(nodesWithinBounds);

        tryInitializeMap();
        addNodesToMap();
    }

    @FXML
    private void selectNode() {
        result.set(nodesComboBox.getValue());
        hideDialog();
    }

    public SparqlService.NodeInfo getResult() {
        return result.getValue();
    }

    public ObjectProperty<SparqlService.NodeInfo> resultProperty() {
        return result;
    }

    public void showDialog(){
        dialogStage.show();
    }

    public void hideDialog(){
        dialogStage.hide();
    }

    public Stage getDialogStage() {
        return dialogStage;
    }
}
