package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.rspec_source.AdvertisementRspecSource;
import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.property.*;
import javafx.beans.value.ObservableListValue;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

/**
 * AuthorityInfo wraps the jFed library's SfaAuthority class with JavaFX properties
 * <p>
 * It also stores additional info related to the current session, such as the last received advertisement RSpec
 * <p>
 * It also offers a convenient and save delete method.
 */
public class AuthorityInfo {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(AuthorityInfo.class);

    private final AuthorityList authorityList;
    private final SfaAuthority geniAuthority;
    //TODO add processed list of resources
    private final ListProperty<String> allowedCertificateHostnameAliases =
            new SimpleListProperty<>(FXCollections.<String>observableArrayList());
    private final ListProperty<AuthorityUrl> urls =
            new SimpleListProperty<>(FXCollections.<AuthorityUrl>observableArrayList());
    /**
     * true if this AuthorityInfo is the same as the SfaAuthority (geniAuthority)
     */
    private BooleanProperty committed = new SimpleBooleanProperty(true);
    /**
     * true if this AuthorityInfo is the same as on disk
     */
    private BooleanProperty saved = new SimpleBooleanProperty(true);
    private boolean hasPreviouslyCommittedChanges = false;
    private final InvalidationListener changeHandler = new InvalidationListener() {
        @Override
        public void invalidated(Observable observable) {
            boolean same = isSame();
            committed.set(same);

            if (saved.get() && !committed.get())
                saved.set(false);

            if (!saved.get() && !hasPreviouslyCommittedChanges && committed.get())
                saved.set(true);
        }
    };
    private int index;
    private StringProperty name = new SimpleStringProperty();
    private StringProperty urn = new SimpleStringProperty("");

    //////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////
    private StringProperty urnPart = new SimpleStringProperty("");
    private StringProperty pemSslTrustCert = new SimpleStringProperty("");

    //////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////
    private BooleanProperty reconnectEachTime = new SimpleBooleanProperty(false);
    private ObjectProperty<AdvertisementRspecSource> availableAdvertisementRspec = new SimpleObjectProperty<>();
    private ObjectProperty<AdvertisementRspecSource> allAdvertisementRspec = new SimpleObjectProperty<>();

    /**
     * Note: typically, you do not want to construct yourself.
     * instead get AuthorityInfo using SfaModel and EasyAuthorityList
     */
    public AuthorityInfo(@Nonnull AuthorityList authorityList, @Nonnull SfaAuthority geniAuthority, int index) {
        this.authorityList = authorityList;
        this.geniAuthority = geniAuthority;
        this.index = index;

        restore();

        //listeners to keep committed and saved properties up to date automatically
        name.addListener(changeHandler);
        urn.addListener(changeHandler);
        //urnPart.addListener(changeHandler); //derived, so not needed to listen to this
        pemSslTrustCert.addListener(changeHandler);
        reconnectEachTime.addListener(changeHandler);
        allowedCertificateHostnameAliases.addListener(changeHandler);
        urls.addListener(changeHandler);
        urls.addListener(new ListChangeListener<AuthorityUrl>() {
            @Override
            public void onChanged(Change<? extends AuthorityUrl> change) {
                while (change.next()) {
                    for (AuthorityUrl added : change.getAddedSubList()) {
                        added.urlProperty().addListener(changeHandler);
                        added.roleProperty().addListener(changeHandler);
                        added.versionProperty().addListener(changeHandler);
                    }
                    for (AuthorityUrl removed : change.getRemoved()) {
                        removed.urlProperty().removeListener(changeHandler);
                        removed.roleProperty().removeListener(changeHandler);
                        removed.versionProperty().removeListener(changeHandler);
                    }
                }
            }
        });
    }

    private void updateUrnPart() {
        String part = "<ILLEGAL URN>";
        try {
            part = SfaAuthority.urnPartFromUrn(urn.get());
        } catch (JFedException ignored) {
        }
        urnPart.set(part);
    }

    public
    @Nonnull
    SfaAuthority getSfaAuthority() {
        return geniAuthority;
    }

    public int getIndex() {
        return index;
    }

    public void setIndex(int index) {
        this.index = index;
    }

    public BooleanProperty committedProperty() {
        return committed;
    }


    //////////////////////////////////////////////////////////////////
    ///////////////////////// observable properties //////////////////
    //////////////////////////////////////////////////////////////////

    public BooleanProperty savedProperty() {
        return saved;
    }

    /**
     * safely delete this from AuthorityListModel and AuthorityList
     */
    public void delete() {
        hasPreviouslyCommittedChanges = true;
        authorityList.getAuthorityListModel().removeByUrn(geniAuthority.getUrnString());
        saved.set(false);
    }

    public void commit() {
        if (isSame()) return;

        geniAuthority.setName(name.get());
//        geniAuthority.setUrn(urn.get());
        geniAuthority.setReconnectEachTime(reconnectEachTime.get());
        geniAuthority.clearPemSslTrustCerts();
        geniAuthority.addPemSslTrustCert(pemSslTrustCert.get());
        geniAuthority.setAllowedCertificateHostnameAlias(allowedCertificateHostnameAliases.get());
        Map<ServerType, URL> newMap = new HashMap<>();
        for (AuthorityUrl authUrl : urls)
            newMap.put(authUrl.getServerType(), authUrl.getURL());
        geniAuthority.replaceUrls(newMap);

        assert isSame();
        committed.set(true);
        hasPreviouslyCommittedChanges = true;
    }

    public void commitAsNew() {
        try {
            SfaAuthority newAuth = new SfaAuthority(urn.get());
            newAuth.setName(name.get());
            newAuth.setReconnectEachTime(reconnectEachTime.get());
            newAuth.clearPemSslTrustCerts();
            newAuth.addPemSslTrustCert(pemSslTrustCert.get());
            newAuth.setAllowedCertificateHostnameAlias(allowedCertificateHostnameAliases.get());
            Map<ServerType, URL> newMap = new HashMap<>();
            for (AuthorityUrl authUrl : urls)
                newMap.put(authUrl.getServerType(), authUrl.getURL());
            newAuth.replaceUrls(newMap);
            hasPreviouslyCommittedChanges = true;
            committed.set(true);
            saved.set(false);
        } catch (JFedException e) {
            e.printStackTrace();
        }
    }

    public void restore() {
        setTo(geniAuthority);
    }

    public void setTo(SfaAuthority sfaAuth) {
        if (sfaAuth == geniAuthority && this.urn != null && this.urn.get() != null) {
            if (isSame()) return;

            //no need for this, changeHandler will automatically do this
//            //because of restore, it is still saved if no changes committed before
//            if (!hasPreviouslyCommittedChanges)
//                saved.set(true);
//
//            committed.set(true);
        }

        //TODO disable changeHandler until the end? (needs to fire only at the end, intermediate changes are not important)

        assert sfaAuth != null;

        this.name.set(sfaAuth.getName());
        this.urn.set(sfaAuth.getUrnString());
        this.urls.clear();
        for (Map.Entry<ServerType, URL> entry : sfaAuth.getUrls().entrySet()) {
            AuthorityUrl authorityUrl = new AuthorityUrl(entry.getKey(), entry.getValue());
            this.urls.add(authorityUrl);
        }
        this.reconnectEachTime.set(sfaAuth.isReconnectEachTime());
        String certs = "";
        if (geniAuthority.getPemSslTrustCerts() != null)
            for (String pem : sfaAuth.getPemSslTrustCerts())
                certs += pem + "\n";
        this.pemSslTrustCert.set(certs);
        allowedCertificateHostnameAliases.clear();
        allowedCertificateHostnameAliases.addAll(sfaAuth.getAllowedCertificateHostnameAliases());

        urn.addListener((observable) -> updateUrnPart());
        updateUrnPart();
    }

    //helper
    @SuppressWarnings("StringEquality")
    private boolean nullSupportingEquals(String a, String b) {
        if (a == null || b == null)
            return a == b;
        return a.equals(b);
    }

    /**
     * return true if this AuthorityInfo is the same as the SfaAuthority (geniAuthority)
     */
    public boolean isSame() {
        assert this.urn != null;
        assert this.urn.get() != null;

        if (!nullSupportingEquals(this.name.get(), geniAuthority.getName())) return false;
        if (!nullSupportingEquals(this.urn.get(), geniAuthority.getUrnString())) return false;
        String certs = "";
        if (geniAuthority.getPemSslTrustCerts() != null)
            for (String pem : geniAuthority.getPemSslTrustCerts())
                certs += pem + "\n";
        if (!nullSupportingEquals(this.pemSslTrustCert.get(), certs)) return false;
        if (this.reconnectEachTime.get() != geniAuthority.isReconnectEachTime()) return false;

        if (this.urls.size() != geniAuthority.getUrls().size()) return false;
        for (AuthorityUrl authorityUrl : this.urls) {
            URL u = geniAuthority.getUrls().get(authorityUrl.getServerType());
            if (u == null) return false;
            if (!u.equals(authorityUrl.getURL())) return false;
        }

        if (this.allowedCertificateHostnameAliases.size() != geniAuthority.getAllowedCertificateHostnameAliases().size())
            return false;
        if (!this.allowedCertificateHostnameAliases.get().containsAll(geniAuthority.getAllowedCertificateHostnameAliases()))
            return false;
        if (!geniAuthority.getAllowedCertificateHostnameAliases().containsAll(this.allowedCertificateHostnameAliases.get()))
            return false;

        return true;
    }

    public void markSaved() {
        if (isSame()) {
            hasPreviouslyCommittedChanges = false;
            saved.set(true);
        } else {
            hasPreviouslyCommittedChanges = false;
        }
    }

    public StringProperty nameProperty() {
        return name;
    }

    public StringProperty urnProperty() {
        return urn;
    }

    public ObservableValue<String> urnPartProperty() {
        return urnPart;
    }

    public StringProperty pemSslTrustCertProperty() {
        return pemSslTrustCert;
    }

    public BooleanProperty reconnectEachTimeProperty() {
        return reconnectEachTime;
    }

    public ObservableListValue<String> getAllowedCertificateHostnameAliases() {
        return allowedCertificateHostnameAliases;
    }

    @Override
    public String toString() {
        return name.get();
    }


    //////////////////////////////////////////////////////////////////
    /////////////////////////// Advertisement RSpec //////////////////
    //////////////////////////////////////////////////////////////////

    public ObservableListValue<AuthorityUrl> getUrls() {
        return urls;
    }

    /**
     * @param isAvailableRspec true if Rspec lists only the available nodes, false if it lists all available nodes
     * @param rspecString RSpec that must be seen
     */
    public void setAdvertisementRspec(boolean isAvailableRspec, @Nullable String rspecString) {
        LOG.trace("AuthorityInfo.seeAdvertisementRspec({}, rspec.length={})",
                isAvailableRspec,
                rspecString != null ? rspecString.length() : -1);

        AdvertisementRspecSource r;
        if (rspecString == null)
            r = null;
        else
            r = new AdvertisementRspecSource(rspecString, ModelRspecType.FX);

        if (isAvailableRspec)
            availableAdvertisementRspec.set(r);
        else
            allAdvertisementRspec.set(r);
    }

    public AdvertisementRspecSource getAvailableAdvertisementRspec() {
        return availableAdvertisementRspec.get();
    }

    public AdvertisementRspecSource getAllAdvertisementRspec() {
        return allAdvertisementRspec.get();
    }

    public ObjectProperty<AdvertisementRspecSource> availableAdvertisementRspecProperty() {
        return availableAdvertisementRspec;
    }

    public ObjectProperty<AdvertisementRspecSource> allAdvertisementRspecProperty() {
        return allAdvertisementRspec;
    }

    public String getName() {
        return name.get();
    }

    //////////////////////////////////////////////////////////////////
    ///////////////////////// observable urls ////////////////////////
    //////////////////////////////////////////////////////////////////
    public static class AuthorityUrl {
        private ObjectProperty<ServerType.GeniServerRole> role = new SimpleObjectProperty<>();
        private StringProperty version = new SimpleStringProperty();
        private ObjectProperty<URL> url = new SimpleObjectProperty<>();

        public AuthorityUrl(ServerType serverType, URL url) {
            this.role.set(serverType.getRole());
            this.version.set(serverType.getVersion());
            this.url.set(url);
        }

        public ServerType getServerType() {
            return new ServerType(role.get(), version.get());
        }

        public URL getURL() {
            return url.get();
        }

        public ObjectProperty<ServerType.GeniServerRole> roleProperty() {
            return role;
        }

        public StringProperty versionProperty() {
            return version;
        }

        public ObjectProperty<URL> urlProperty() {
            return url;
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof AuthorityInfo)) return false;

        AuthorityInfo that = (AuthorityInfo) o;

        if (!geniAuthority.equals(that.geniAuthority))
            return false;

        return true;
    }

    @Override
    public int hashCode() {
        return geniAuthority.hashCode();
    }
}
